@@ -16,12 +16,84 @@ os.environ["VLLM_WORKER_MULTIPROC_METHOD"] = "spawn"
 
 
 def pytest_collection_modifyitems(config, items):
-    """ Mark all tests in e2e directory"""
+    """ Modify tests at collection time """
+    _mark_all_e2e(items)
+
+    _skip_quantized_by_default(config, items)
+
+    _xfail_fp8_on_spyre(items)
+
+    _skip_all_cb_and_fp8_tests(items)
+
+    _skip_unsupported_compiler_tests(config, items)
+
+
+def _mark_all_e2e(items):
+    """Mark all tests within the e2e package with the e2e marker"""
     for item in items:
         if "e2e" in str(item.nodeid):
             item.add_marker(pytest.mark.e2e)
 
 
+def _skip_quantized_by_default(config, items):
+    """Skip tests marked with `quantized` unless the `-m` flag includes it
+    Ref: https://stackoverflow.com/questions/56374588/how-can-i-ensure-tests-with-a-marker-are-only-run-if-explicitly-asked-in-pytest
+
+    This will skip the quantized tests at runtime, but they will still show up
+    as collected when running pytest --collect-only.
+    """
+    markexpr = config.option.markexpr
+    if "quantized" in markexpr:
+        return  # let pytest handle the collection logic
+
+    skip_mymarker = pytest.mark.skip(reason='quantized not selected')
+    for item in items:
+        if "quantized" in item.keywords:
+            item.add_marker(skip_mymarker)
+
+
+def _xfail_fp8_on_spyre(items):
+    """Set an xfail marker on all tests that run quantized models on Spyre
+    hardware.
+    
+    TODO: Relax this to only "spyre and cb" once static batching is supported
+    on spyre.
+    """
+
+    xfail_marker = pytest.mark.xfail(
+        reason="fp8 is not yet supported on Spyre")
+    for item in items:
+        if "quantized" in item.keywords and "spyre" in item.keywords:
+            item.add_marker(xfail_marker)
+
+
+def _skip_all_cb_and_fp8_tests(items):
+    """Skip all tests that run fp8 with continuous batching.
+    This can be relaxed once the TODOs to implement fp8 paged attention are
+    resolved.
+    """
+    skip_marker = pytest.mark.skip(
+        reason="FP8 is not supported with continuous batching yet")
+    for item in items:
+        if "quantized" in item.keywords and "cb" in item.keywords:
+            item.add_marker(skip_marker)
+
+
+def _skip_unsupported_compiler_tests(config, items):
+    """Skip all tests that need compiler changes to run.
+    This can be relaxed once the compiler changes are in place
+    """
+
+    markexpr = config.option.markexpr
+    if "compiler_support_16k" in markexpr:
+        return  # let pytest handle the collection logic
+
+    skip_marker = pytest.mark.skip(reason="Needs compiler changes")
+    for item in items:
+        if "spyre" in item.keywords and "compiler_support_16k" in item.keywords:
+            item.add_marker(skip_marker)
+
+
 @pytest.fixture(autouse=True)
 def init_test_http_connection():
     # pytest_asyncio may use a different event loop per test
@@ -79,19 +151,29 @@ def remote_openai_server(request):
         raise pytest.UsageError(
             "Error setting up remote_openai_server params") from e
 
-    if 'cb' in params:
+    # Default to None if not present
+    quantization = params.get("quantization", None)
+
+    # Add extra server args if present in test
+    server_args = ["--quantization", quantization] if quantization else []
+
+    if 'tp_size' in params:
+        tp_size = params['tp_size']
+        skip_unsupported_tp_size(int(tp_size), backend)
+        server_args.extend(["--tensor-parallel-size", str(tp_size)])
+
+    if "cb" in params and params["cb"] == 1:
         max_model_len = params["max_model_len"]
         max_num_seqs = params["max_num_seqs"]
         env_dict = {
             "VLLM_SPYRE_USE_CB": "1",
             "VLLM_SPYRE_DYNAMO_BACKEND": backend
         }
-        server_args = [
+        server_args.extend([
             "--max_num_seqs",
             str(max_num_seqs), "--max-model-len",
             str(max_model_len)
-        ]
-
+        ])
     else:
         warmup_shape = params['warmup_shape']
         warmup_prompt_length = [t[0] for t in warmup_shape]
@@ -108,17 +190,6 @@ def remote_openai_server(request):
             backend,
         }
 
-        # Default to None if not present
-        quantization = params.get('quantization', None)
-
-        # Add extra server args if present in test
-        server_args = ["--quantization", quantization] if quantization else []
-
-        if 'tp_size' in params:
-            tp_size = params['tp_size']
-            skip_unsupported_tp_size(int(tp_size), backend)
-            server_args.extend(["--tensor-parallel-size", str(tp_size)])
-
     try:
         with RemoteOpenAIServer(model, server_args,
                                 env_dict=env_dict) as server:
