@@ -33,7 +33,7 @@ def test_prompts_aligned_with_tkv_boundaries(model: str, backend: str,
     seqs_max_tokens = [65, 67, 7]
     prompts_lengths = [49, 41, 47]
     steps_add_reqs = [0, 0, 0]  # add all requests in the beginning
-    available_blocks = -1  # no restriction
+    available_blocks = 1000  # no restriction
     max_num_seqs = 2
     max_model_len = 256
 
@@ -202,7 +202,7 @@ def test_prompts_misaligned_with_tkv_boundaries(
     seqs_max_tokens = [57, 67, 9]
     prompts_lengths = [49, 41, 47]
     steps_add_reqs = [0, 0, 0]  # add all requests in the beginning
-    available_blocks = -1  # no restriction
+    available_blocks = 1000  # no restriction
     max_num_seqs = 2
     max_model_len = 256
 
@@ -368,7 +368,7 @@ def test_two_sequences_finish_same_time_as_new_arrive(
     seqs_max_tokens = [30, 30, 10]
     prompts_lengths = [49, 30, 20]
     steps_add_reqs = [0, 0, 31]
-    available_blocks = -1  # no restriction
+    available_blocks = 1000  # no restriction
     max_num_seqs = 2
     max_model_len = 256
 
@@ -513,7 +513,7 @@ def test_new_sequence_joins_during_decode(model: str, backend: str,
     seqs_max_tokens = [119, 52, 104, 64]
     prompts_lengths = [49, 14, 89, 9]
     steps_add_reqs = [0, 0, 32, 131]
-    available_blocks = -1  # no restriction
+    available_blocks = 1000  # no restriction
     max_num_seqs = 4
     max_model_len = 256
 
@@ -764,10 +764,15 @@ def test_new_sequence_joins_during_decode(model: str, backend: str,
 @pytest.mark.cb
 @pytest.mark.parametrize("model", get_spyre_model_list())
 @pytest.mark.parametrize("backend", get_spyre_backend_list())
+@pytest.mark.parametrize("prefill_optimization", [True, False])
 def test_prompt_too_long_for_current_tkv(model: str, backend: str,
+                                         prefill_optimization: bool,
                                          monkeypatch: pytest.MonkeyPatch,
                                          set_random_seed):
     """ Scenario where the requested prompt is too long for current tkv value
+   
+    Note that with VLLM_SPYRE_ENABLE_PREFILL_OPTIMIZATION enabled, we can 
+    prefill the prompt straight away -> using checked_steps_with_optimization
 
     Configuration:
         * max_num_seqs: 2
@@ -776,10 +781,13 @@ def test_prompt_too_long_for_current_tkv(model: str, backend: str,
             * 2: len = 70, max tokens = 67, step joining = 0
     """
 
+    if prefill_optimization:
+        monkeypatch.setenv('VLLM_SPYRE_ENABLE_PREFILL_OPTIMIZATION', '1')
+
     seqs_max_tokens = [57, 67]
     prompts_lengths = [49, 70]
     steps_add_reqs = [0, 0]
-    available_blocks = -1  # no restriction
+    available_blocks = 1000  # no restriction
     max_num_seqs = 2
     max_model_len = 256
 
@@ -896,6 +904,109 @@ def test_prompt_too_long_for_current_tkv(model: str, backend: str,
         },
     ]
 
+    checked_steps_with_optimization = [
+        {
+            "step": 0,
+            "tkv": 0,
+            "waiting": ["0", "1"],
+            "running": [],
+            "request_outputs": [],
+            "n_reserved_blocks": 0,
+            "n_used_blocks": 0
+        },
+        {
+            # Prefill sequence 0
+            # total blocks in use: 1
+            "step": 1,
+            "tkv": 64,
+            "waiting": ["1"],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks": 2,  # prefill (1 block) + 56 decodes (1 block)
+            "n_used_blocks": 1
+        },
+        # due to allowing sequences to join the current decode batch even if
+        # prompt length > tkv, prefill of sequence 1 happens immediately
+        {
+            # Prefill sequence 1
+            # total blocks in use: 2 + 2
+            "step": 2,
+            "tkv": 128,
+            "waiting": [],
+            "running": ["1", "0"],
+            "request_outputs": ["1"],
+            # 2 + 4 (prefill (2 block) + 66 decodes (2 blocks))
+            "n_reserved_blocks": 6,
+            "n_used_blocks": 3
+        },
+        {
+            # Decode sequences 0 and 1
+            "step": 3,
+            "tkv": 129,
+            "waiting": [],
+            "running": ["1", "0"],
+            "request_outputs": ["1", "0"],
+            "n_reserved_blocks": 6,
+            "n_used_blocks": 5  # 3 + 2 = 5
+        },
+        {
+            # Sequence 0 finishes at step 58
+            # (start step + 2 prefills + 56 decodes - 1) = 1 + 2 + 56 - 1 = 58
+            "step": 58,
+            "tkv": 184,
+            "waiting": [],
+            "running": ["1"],
+            "request_outputs": ["1", "0"],
+            "finished_requests": ["0"],
+            "n_reserved_blocks": 6,
+            "n_used_blocks": 5
+        },
+        {
+            # Decode sequence 1
+            # total blocks in use: 5 - 2 = 3
+            "step": 59,
+            "tkv": 185,
+            "waiting": [],
+            "running": ["1"],
+            "request_outputs": ["1"],
+            "n_reserved_blocks": 4,  # 6 - 2 (seq 0)
+            "n_used_blocks": 3
+        },
+        {
+            # Decode sequence 1 needs another block
+            # total blocks in use: 3 + 1 = 4
+            "step": 67,
+            "tkv": 193,
+            "waiting": [],
+            "running": ["1"],
+            "request_outputs": ["1"],
+            "n_reserved_blocks": 4,
+            "n_used_blocks": 4
+        },
+        {
+            # Sequence 1 finishes at step 74
+            # (start step + 1 prefill + 66 decodes - 1) = 2 + 1 + 66 - 1 = 68
+            "step": 68,
+            "tkv": 194,
+            "waiting": [],
+            "running": [],
+            "request_outputs": ["1"],
+            "finished_requests": ["1"],
+            "n_reserved_blocks": 4,
+            "n_used_blocks": 4
+        },
+        {
+            # Tkv should be cleared one step later
+            "step": 69,
+            "tkv": 0,
+            "waiting": [],
+            "running": [],
+            "request_outputs": [],
+            "n_reserved_blocks": 0,
+            "n_used_blocks": 0
+        },
+    ]
+
     cb_outputs, prompts = check_scheduler_inference_steps(
         model=model,
         backend=backend,
@@ -903,7 +1014,347 @@ def test_prompt_too_long_for_current_tkv(model: str, backend: str,
         seqs_max_tokens=seqs_max_tokens,
         prompts_lengths=prompts_lengths,
         steps_add_reqs=steps_add_reqs,
-        checked_steps=checked_steps,
+        checked_steps=checked_steps_with_optimization
+        if prefill_optimization else checked_steps,
+        max_num_seqs=max_num_seqs,
+        max_model_len=max_model_len,
+        available_blocks=available_blocks,
+        use_cb=True,
+    )
+
+    check_output_against_hf(model, backend, seqs_max_tokens, cb_outputs,
+                            prompts)
+
+
+@pytest.mark.cb
+@pytest.mark.parametrize("model", get_spyre_model_list())
+@pytest.mark.parametrize("backend", get_spyre_backend_list())
+def test_prefill_optimization_tkv_too_big(model: str, backend: str,
+                                          monkeypatch: pytest.MonkeyPatch,
+                                          set_random_seed):
+    """ Scenario where the requested prompt is too long for current tkv value
+   
+    Note that as VLLM_SPYRE_ENABLE_PREFILL_OPTIMIZATION is enabled, we could 
+    prefill the prompt straight away -> using checked_steps_with_optimization
+
+    However, in this test the max model length is decreased to a value where
+    the tkv of the decode batch would be shifted beyond the max model length, 
+    we therefore have to wait with scheduling it via the prefill optimization. 
+    -> see cond4_updated in vllm_spyre/v1/core/scheduler.py
+
+    Configuration:
+        * max_num_seqs: 2
+        * number of prompts: 2
+            * 1: len = 49, max tokens = 67, step joining = 0
+            * 2: len = 70, max tokens = 50, step joining = 0
+    """
+
+    monkeypatch.setenv('VLLM_SPYRE_ENABLE_PREFILL_OPTIMIZATION', '1')
+
+    seqs_max_tokens = [67, 50]
+    prompts_lengths = [49, 70]
+    steps_add_reqs = [0, 0]
+    available_blocks = 1000  # no restriction
+    max_num_seqs = 2
+    # restricting the max model length here to trigger the violated
+    # scheduler condition
+    max_model_len = 192
+
+    checked_steps_with_optimization = [
+        {
+            "step": 0,
+            "tkv": 0,
+            "waiting": ["0", "1"],
+            "running": [],
+            "request_outputs": [],
+            "n_reserved_blocks": 0,
+            "n_used_blocks": 0
+        },
+        {
+            # Prefill sequence 0
+            # total blocks in use: 1
+            "step": 1,
+            "tkv": 64,
+            "waiting": ["1"],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks":
+            3,  # prefill (1 block) + 66 decodes (2 blocks)
+            "n_used_blocks": 1
+        },
+        # Here we cannot schedule sequence 1. By shifting sequence 0 by 1 block
+        # due to the prefill optimization, its max tkv would exceed the max
+        # model length: 64 + 67 - 1 + 64 (shift) = 194 > 192 (max model length)
+        {
+            # Decode sequence 0
+            # total blocks in use: 1 + 1
+            "step": 2,
+            "tkv": 65,
+            "waiting": ["1"],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks": 3,
+            "n_used_blocks": 2
+        },
+        {
+            # Prefill sequence 1, tkv large enough to prefill w/o optimization
+            # total blocks in use: 2 + 2
+            "step": 8,
+            "tkv": 70,
+            "waiting": [],
+            "running": ["1", "0"],
+            "request_outputs": ["1"],
+            # 3 + 2 (prefill (2 block) + 49 decodes in the last block)
+            "n_reserved_blocks": 5,
+            "n_used_blocks": 4
+        },
+        {
+            # Decode sequences 0 and 1
+            "step": 9,
+            "tkv": 71,
+            "waiting": [],
+            "running": ["1", "0"],
+            "request_outputs": ["1", "0"],
+            "n_reserved_blocks": 5,
+            "n_used_blocks": 4  # seq 1 writes into the right pads
+        },
+        {
+            # Sequence 1 finishes at step 57
+            # (start step 8 + 1 prefills + 49 decodes - 1) = 8 + 1 + 49 - 1 = 57
+            "step": 57,
+            "tkv": 119,
+            "waiting": [],
+            "running": ["0"],
+            "request_outputs": ["1", "0"],
+            "finished_requests": ["1"],
+            "n_reserved_blocks": 5,
+            "n_used_blocks": 4
+        },
+        {
+            # Decode sequence 0
+            # total blocks in use: 4 - 2 = 2
+            "step": 58,
+            "tkv": 120,
+            "waiting": [],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks": 3,  # 5 - 2 (seq 1)
+            "n_used_blocks": 2
+        },
+        {
+            # Decode sequence 0 needs another block
+            # total blocks in use: 2 + 1 = 3
+            "step": 67,
+            "tkv": 129,
+            "waiting": [],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks": 3,
+            "n_used_blocks": 3
+        },
+        {
+            # Sequence 0 finishes at step 68
+            # (start step + 2 prefill + 66 decodes - 1) = 1 + 2 + 66 - 1 = 68
+            "step": 68,
+            "tkv": 130,
+            "waiting": [],
+            "running": [],
+            "request_outputs": ["0"],
+            "finished_requests": ["0"],
+            "n_reserved_blocks": 3,
+            "n_used_blocks": 3
+        },
+        {
+            # Tkv should be cleared one step later
+            "step": 69,
+            "tkv": 0,
+            "waiting": [],
+            "running": [],
+            "request_outputs": [],
+            "n_reserved_blocks": 0,
+            "n_used_blocks": 0
+        },
+    ]
+
+    cb_outputs, prompts = check_scheduler_inference_steps(
+        model=model,
+        backend=backend,
+        monkeypatch=monkeypatch,
+        seqs_max_tokens=seqs_max_tokens,
+        prompts_lengths=prompts_lengths,
+        steps_add_reqs=steps_add_reqs,
+        checked_steps=checked_steps_with_optimization,
+        max_num_seqs=max_num_seqs,
+        max_model_len=max_model_len,
+        available_blocks=available_blocks,
+        use_cb=True,
+    )
+
+    check_output_against_hf(model, backend, seqs_max_tokens, cb_outputs,
+                            prompts)
+
+
+@pytest.mark.cb
+@pytest.mark.parametrize("model", get_spyre_model_list())
+@pytest.mark.parametrize("backend", get_spyre_backend_list())
+def test_prefill_optimization_use_more_than_available_blocks(
+        model: str, backend: str, monkeypatch: pytest.MonkeyPatch,
+        set_random_seed):
+    """ Scenario where the requested prompt is too long for current tkv value
+   
+    Note that as VLLM_SPYRE_ENABLE_PREFILL_OPTIMIZATION is enabled, we could 
+    prefill the prompt straight away -> using checked_steps_with_optimization
+
+    However, in this test the number of available KV cache blocks is decreased
+    to a value where the the number of reserved blocks would exceed the number
+    of available blocks, we therefore have to wait with scheduling it via the 
+    prefill optimization. 
+    -> see cond5_updated in vllm_spyre/v1/core/scheduler.py
+
+    Configuration:
+        * max_num_seqs: 2
+        * number of prompts: 2
+            * 1: len = 49, max tokens = 67, step joining = 0
+            * 2: len = 70, max tokens = 50, step joining = 0
+    """
+
+    monkeypatch.setenv('VLLM_SPYRE_ENABLE_PREFILL_OPTIMIZATION', '1')
+
+    seqs_max_tokens = [67, 50]
+    prompts_lengths = [49, 70]
+    steps_add_reqs = [0, 0]
+    # provide only 5 blocks, to use the prefill optimization
+    # at least 6 blocks would be required
+    available_blocks = 5
+    max_num_seqs = 2
+    max_model_len = 256
+
+    checked_steps_with_optimization = [
+        {
+            "step": 0,
+            "tkv": 0,
+            "waiting": ["0", "1"],
+            "running": [],
+            "request_outputs": [],
+            "n_reserved_blocks": 0,
+            "n_used_blocks": 0
+        },
+        {
+            # Prefill sequence 0
+            # total blocks in use: 1
+            "step": 1,
+            "tkv": 64,
+            "waiting": ["1"],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks":
+            3,  # prefill (1 block) + 66 decodes (2 blocks)
+            "n_used_blocks": 1
+        },
+        # We cannot schedule sequence 1 here. Prefill optimization shifts
+        # sequence 0 by 1 block, so it still needs 3 blocks (not counting
+        # fully padded blocks!) Aligning sequence 1 would then require
+        # 3 blocks (instead of 2). With only 5 blocks available, scheduling
+        # sequence 1 is not possible.
+        {
+            # Decode sequence 0
+            # total blocks in use: 1 + 1
+            "step": 2,
+            "tkv": 65,
+            "waiting": ["1"],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks": 3,
+            "n_used_blocks": 2
+        },
+        {
+            # Prefill sequence 1, tkv large enough to prefill w/o optimization
+            # total blocks in use: 2 + 2
+            "step": 8,
+            "tkv": 70,
+            "waiting": [],
+            "running": ["1", "0"],
+            "request_outputs": ["1"],
+            # 3 + 2 (prefill (2 block) + 49 decodes in the last block)
+            "n_reserved_blocks": 5,
+            "n_used_blocks": 4
+        },
+        {
+            # Decode sequences 0 and 1
+            "step": 9,
+            "tkv": 71,
+            "waiting": [],
+            "running": ["1", "0"],
+            "request_outputs": ["1", "0"],
+            "n_reserved_blocks": 5,
+            "n_used_blocks": 4  # seq 1 writes into the right pads
+        },
+        {
+            # Sequence 1 finishes at step 57
+            # (start step 8 + 1 prefills + 49 decodes - 1) = 8 + 1 + 49 - 1 = 57
+            "step": 57,
+            "tkv": 119,
+            "waiting": [],
+            "running": ["0"],
+            "request_outputs": ["1", "0"],
+            "finished_requests": ["1"],
+            "n_reserved_blocks": 5,
+            "n_used_blocks": 4
+        },
+        {
+            # Decode sequence 0
+            # total blocks in use: 4 - 2 = 2
+            "step": 58,
+            "tkv": 120,
+            "waiting": [],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks": 3,  # 5 - 2 (seq 1)
+            "n_used_blocks": 2
+        },
+        {
+            # Decode sequence 0 needs another block
+            # total blocks in use: 2 + 1 = 3
+            "step": 67,
+            "tkv": 129,
+            "waiting": [],
+            "running": ["0"],
+            "request_outputs": ["0"],
+            "n_reserved_blocks": 3,
+            "n_used_blocks": 3
+        },
+        {
+            # Sequence 0 finishes at step 68
+            # (start step + 2 prefill + 66 decodes - 1) = 1 + 2 + 66 - 1 = 68
+            "step": 68,
+            "tkv": 130,
+            "waiting": [],
+            "running": [],
+            "request_outputs": ["0"],
+            "finished_requests": ["0"],
+            "n_reserved_blocks": 3,
+            "n_used_blocks": 3
+        },
+        {
+            # Tkv should be cleared one step later
+            "step": 69,
+            "tkv": 0,
+            "waiting": [],
+            "running": [],
+            "request_outputs": [],
+            "n_reserved_blocks": 0,
+            "n_used_blocks": 0
+        },
+    ]
+
+    cb_outputs, prompts = check_scheduler_inference_steps(
+        model=model,
+        backend=backend,
+        monkeypatch=monkeypatch,
+        seqs_max_tokens=seqs_max_tokens,
+        prompts_lengths=prompts_lengths,
+        steps_add_reqs=steps_add_reqs,
+        checked_steps=checked_steps_with_optimization,
         max_num_seqs=max_num_seqs,
         max_model_len=max_model_len,
         available_blocks=available_blocks,
@@ -933,7 +1384,7 @@ def test_requested_tokens_not_fitting_remaining_space(
     seqs_max_tokens = [67, 57, 80]
     prompts_lengths = [70, 49, 41]
     steps_add_reqs = [0, 0, 0]
-    available_blocks = -1  # no restriction
+    available_blocks = 1000  # no restriction
     max_num_seqs = 2
     max_model_len = 256
 
