@@ -1,3 +1,4 @@
+import inspect
 import math
 import os
 import random
@@ -12,18 +13,24 @@ import openai
 import pytest
 import requests
 import torch
+from hf_result_cache import HFResultCache
 from sentence_transformers import SentenceTransformer, util
 from transformers import AutoModelForCausalLM, AutoTokenizer
 from vllm import LLM, SamplingParams
 from vllm.entrypoints.openai.cli_args import make_arg_parser
 from vllm.utils import FlexibleArgumentParser, get_open_port
 from vllm.v1.engine import EngineCoreRequest
+from vllm.v1.engine.core import EngineCore
+from vllm.v1.request import Request
 
 DISABLE_ASSERTS = False  # used for debugging
 
-ISCLOSE_REL_TOL_CPU = 0.2
+# TODO: Needs to be separate for quantized models
+ISCLOSE_REL_TOL_CPU = 0.35
 ISCLOSE_REL_TOL_SPYRE = 0.35
 
+HF_RESULT_CACHE = HFResultCache()
+
 
 def force_engine_shutdown(llm: LLM):
     """
@@ -171,12 +178,29 @@ def patch_warmup_shapes(warmup_shapes: Union[list[tuple[int, int, int]],
                            ','.join(str(val) for val in warmup_new_tokens))
 
 
+def extract_output(req_output):
+    """Extract text, token_ids, tokens, and logprobs from request output."""
+
+    result = {}
+    result['text'] = req_output.outputs[0].text
+
+    # TODO: Workaround for V1, if request does not fit in a warmup shape
+    # token_ids may be filled with -1.
+    token_ids = [t for t in req_output.outputs[0].token_ids if t >= 0]
+    result['token_ids'] = tuple(token_ids)
+    result['tokens'] = tuple(req_output.outputs[0].logprobs[i][t].decoded_token
+                             for i, t in enumerate(token_ids))
+    result['logprobs'] = tuple(req_output.outputs[0].logprobs[i][t].logprob
+                               for i, t in enumerate(token_ids))
+
+    return result
+
+
 # vLLM / Spyre
 def generate_spyre_vllm_output(
     model: str,
     prompts: Union[list[str], list[list[int]]],
     max_model_len: int,
-    block_size: int,
     sampling_params: Union[SamplingParams, list[SamplingParams]],
     tensor_parallel_size: int,
     backend: str,
@@ -214,7 +238,6 @@ def generate_spyre_vllm_output(
         tokenizer=model,
         max_model_len=max_model_len,
         max_num_seqs=max_num_seqs,
-        block_size=block_size,
         tensor_parallel_size=tensor_parallel_size,
     )
 
@@ -222,20 +245,7 @@ def generate_spyre_vllm_output(
     results = []
 
     for req_output in vllm_outputs:
-        result = {}
-        result['text'] = req_output.outputs[0].text
-        # TODO: Workaround for V1, if request does not fit in a warmup shape
-        # token_ids may be filled with -1.
-        token_ids = [t for t in req_output.outputs[0].token_ids if t >= 0]
-        result['token_ids'] = tuple(token_ids)
-        result['tokens'] = tuple([
-            req_output.outputs[0].logprobs[i][t].decoded_token
-            for i, t in enumerate(result['token_ids'])
-        ])
-        result['logprobs'] = tuple([
-            req_output.outputs[0].logprobs[i][t].logprob
-            for i, t in enumerate(result['token_ids'])
-        ])
+        result = extract_output(req_output)
         results.append(result)
 
     force_engine_shutdown(vllm_model)
@@ -244,22 +254,37 @@ def generate_spyre_vllm_output(
 
 # Hugging Face
 def generate_hf_output(
-    model: str,
-    prompts: Union[list[str], list[list[int]]],  # also accept token ids
-    max_new_tokens: Union[int, list[int]],
-    ignore_eos: bool = False,
-) -> list[dict[str, Any]]:
+        model: str,
+        prompts: Union[list[str], list[list[int]]],  # also accept token ids
+        max_new_tokens: Union[int, list[int]],
+        ignore_eos: bool = False,
+        include_prompt: bool = False) -> list[dict[str, Any]]:
+    """Loads and runs the model on cpu with transformers, caching the results.
+    Returns cached results if any are found to avoid overhead."""
 
     if not isinstance(max_new_tokens, list):
         max_new_tokens = [max_new_tokens] * len(prompts)
 
+    results = []
+    for prompt, max_tokens in zip(prompts, max_new_tokens):
+        results.append(
+            HF_RESULT_CACHE.get_cached_result(model, prompt, max_tokens))
+
+    if all(results):
+        # Everything hit cache
+        return results
+
     hf_model = AutoModelForCausalLM.from_pretrained(model)
     hf_tokenizer = AutoTokenizer.from_pretrained(model)
     if ignore_eos:
         hf_model.generation_config.eos_token_id = None
 
-    results = []
     for prompt_index, prompt in enumerate(prompts):
+
+        if results[prompt_index]:
+            # Already have cached result
+            continue
+
         hf_input_tokens = hf_tokenizer(prompt, return_tensors="pt").input_ids \
                     if isinstance(prompt[0], str) \
                     else torch.tensor([prompts[prompt_index]])
@@ -291,8 +316,16 @@ def generate_hf_output(
         result['token_ids'] = tuple(result['token_ids'])
         result['tokens'] = tuple(result['tokens'])
         result['logprobs'] = tuple(result['logprobs'])
-        results.append(result)
+        if include_prompt:
+            result['prompt'] = prompt
 
+        # Save and cache new result
+        results[prompt_index] = result
+        HF_RESULT_CACHE.add_to_cache(model, prompt,
+                                     max_new_tokens[prompt_index], result)
+
+    # Write back to the cache
+    HF_RESULT_CACHE.write_cache()
     return results
 
 
@@ -332,8 +365,13 @@ def compare_results(
             print(f"        vLLM:  {repr(vllm_result['text']):s}{err_msg}")
             print()
 
+        if isinstance(hf_result['token_ids'], list):
+            hf_result['token_ids'] = tuple(hf_result['token_ids'])
+
         assert DISABLE_ASSERTS or backend == 'sendnn' or\
-            hf_result['token_ids'] == vllm_result['token_ids']
+            hf_result['token_ids'] == vllm_result['token_ids'], \
+            f"Token ids differ: {hf_result['token_ids']} != " \
+            f"{vllm_result['token_ids']}"
 
         if len(hf_result['tokens']) > 0:
             print("   token id. token               logprob      "
@@ -415,13 +453,12 @@ def check_output_against_hf(model, backend, max_new_tokens, vllm_results,
 
 # vLLM / Spyre
 def spyre_vllm_embeddings(model: str, prompts: list[str], max_model_len: int,
-                          block_size: int, tensor_parallel_size: int,
+                          tensor_parallel_size: int,
                           backend: str) -> list[dict[str, Any]]:
 
     vllm_model = LLM(model=model,
                      tokenizer=model,
                      max_model_len=max_model_len,
-                     block_size=block_size,
                      tensor_parallel_size=tensor_parallel_size)
 
     vllm_outputs = vllm_model.embed(prompts)
@@ -510,31 +547,14 @@ def get_spyre_backend_list():
 # get model names from env, if not set then use default models for each type.
 # Multiple models can be specified with a comma separated list in
 # VLLM_SPYRE_TEST_MODEL_LIST
-def get_spyre_model_list(isEmbeddings=False, quantization=None):
+def get_spyre_model_list(isEmbeddings=False):
+    user_test_model_list = os.environ.get("VLLM_SPYRE_TEST_MODEL_LIST")
+    if not user_test_model_list:
+        return _default_test_models(isEmbeddings)
+
+    # User overridden model list
     spyre_model_dir_path = get_spyre_model_dir_path()
-
-    def _get_or_default(env: str, default: str) -> str:
-        """Handle empty strings in env var"""
-        val = os.environ.get(env, default)
-        if not val:
-            val = default
-        return val
-
-    if isEmbeddings:
-        user_test_model_list = _get_or_default(
-            "VLLM_SPYRE_TEST_MODEL_LIST",
-            "sentence-transformers/all-roberta-large-v1")
-        marks = [pytest.mark.embedding]
-    elif quantization == "gptq":
-        # TODO: need a HF hub reference here as a default
-        user_test_model_list = _get_or_default("VLLM_SPYRE_TEST_MODEL_LIST",
-                                               "granite-3.0-8b-instruct-gptq")
-        marks = [pytest.mark.decoder, pytest.mark.quantized, pytest.mark.spyre]
-    else:
-        user_test_model_list = _get_or_default(
-            "VLLM_SPYRE_TEST_MODEL_LIST",
-            "ibm-ai-platform/micro-g3.3-8b-instruct-1b")
-        marks = [pytest.mark.decoder]
+    marks = [pytest.mark.embedding] if isEmbeddings else [pytest.mark.decoder]
 
     test_model_list = []
     for model in user_test_model_list.split(","):
@@ -544,7 +564,29 @@ def get_spyre_model_list(isEmbeddings=False, quantization=None):
     return test_model_list
 
 
-def create_text_prompt(model: str, min_tokens: int, max_tokens: int) -> str:
+def _default_test_models(isEmbeddings=False):
+    """Return the default set of test models as pytest parameterizations"""
+    if isEmbeddings:
+        model = "sentence-transformers/all-roberta-large-v1"
+        return [pytest.param(model, marks=[pytest.mark.embedding], id=model)]
+
+    # Decoders
+    # We run tests for both the full-precision bf16 and fp8-quantized models,
+    # but by default the `pytest.mark.quantized` marker is de-selected unless
+    # the test command includes `-m quantized`.
+    tinygranite = "ibm-ai-platform/micro-g3.3-8b-instruct-1b"
+    tinygranite_fp8 = "ibm-ai-platform/micro-g3.3-8b-instruct-1b-FP8"
+    params = [
+        pytest.param(tinygranite, marks=[pytest.mark.decoder], id=tinygranite),
+        pytest.param(tinygranite_fp8,
+                     marks=[pytest.mark.decoder, pytest.mark.quantized],
+                     id=tinygranite_fp8)
+    ]
+    return params
+
+
+def create_text_prompt(model: str, min_token_length: int,
+                       max_token_length: int) -> str:
     """Create a text prompt for the specified model that will tokenize to within
     the specified token length range."""
     tokenizer = AutoTokenizer.from_pretrained(model)
@@ -552,23 +594,48 @@ def create_text_prompt(model: str, min_tokens: int, max_tokens: int) -> str:
     pepper_tokens = len(tokenizer.encode(pepper, add_special_tokens=False))
 
     # Find a good starting number of peppers
-    prompt = pepper * (min_tokens // pepper_tokens + 1)
+    prompt = pepper * (min_token_length // pepper_tokens + 1)
 
     # And add more until we're over the minimum token length
-    while len(tokenizer.encode(prompt)) <= min_tokens:
+    while len(tokenizer.encode(prompt)) <= min_token_length:
         prompt += pepper
 
     # Make sure this prompt is within the specified range
-    assert min_tokens < len(tokenizer.encode(prompt)) < max_tokens
+    assert min_token_length < len(tokenizer.encode(prompt)) < max_token_length
 
     return prompt
 
 
-def create_random_request(request_id: int,
-                          num_tokens: int,
-                          sampling_params: SamplingParams,
-                          from_model_vocab: bool = False,
-                          model: Optional[str] = None) -> EngineCoreRequest:
+def create_seq_prompt(model: str, token_length: int) -> str:
+    """Create a repeating sequential number prompt for the specified
+    model that will tokenize to exactly the specified token length."""
+
+    tokenizer = AutoTokenizer.from_pretrained(model)
+
+    # 20-token pattern
+    pattern = "0 1 2 3 4 5 6 7 8 9 "
+
+    # Repeat to token_length
+    repeat_count = (token_length // 20) + 1
+    text_prompt = pattern * repeat_count
+
+    # Tokenize and slice
+    tokens = tokenizer.encode(text_prompt)[:token_length]
+
+    # Assert exact token length
+    assert len(tokens) == token_length, \
+        f"Token length mismatch: {len(tokens)} != {token_length}"
+
+    return tokenizer.decode(tokens)
+
+
+def create_random_request(
+    request_id: int,
+    num_tokens: int,
+    sampling_params: SamplingParams,
+    from_model_vocab: bool = False,
+    model: Optional[str] = None,
+) -> Request | EngineCoreRequest:
 
     tokenizer = AutoTokenizer.from_pretrained(model)
     if from_model_vocab:
@@ -590,19 +657,39 @@ def create_random_request(request_id: int,
         assert (len(prompt_token_ids) == num_tokens
                 ), f"need {num_tokens} but got {len(prompt_token_ids)}"
 
-    return EngineCoreRequest(
+    sig = inspect.signature(EngineCore.add_request)
+    inputs_renamed = hasattr(EngineCoreRequest, 'mm_kwargs')
+    if sig.parameters["request"].annotation == EngineCoreRequest:
+        kwargs = {"mm_kwargs" if inputs_renamed else "mm_inputs": None}
+        return EngineCoreRequest(
+            request_id=str(request_id),
+            prompt_token_ids=prompt_token_ids,
+            mm_hashes=None,
+            mm_placeholders=None,
+            sampling_params=sampling_params,
+            eos_token_id=None,
+            arrival_time=0,
+            lora_request=None,
+            data_parallel_rank=None,
+            pooling_params=None,
+            cache_salt=None,
+            **kwargs,
+        )
+    kwargs = {
+        "multi_modal_kwargs" if inputs_renamed else "multi_modal_inputs": None
+    }
+    return Request(
         request_id=str(request_id),
         prompt_token_ids=prompt_token_ids,
-        mm_inputs=None,
-        mm_hashes=None,
-        mm_placeholders=None,
+        multi_modal_hashes=None,
+        multi_modal_placeholders=None,
         sampling_params=sampling_params,
         eos_token_id=None,
         arrival_time=0,
         lora_request=None,
-        data_parallel_rank=None,
         pooling_params=None,
         cache_salt=None,
+        **kwargs,
     )
 
 
@@ -677,3 +764,48 @@ def get_longer_chicken_soup_prompts(num_prompts: int) -> list[str]:
         prompts = prompts * (math.ceil(num_prompts / 4))
 
     return prompts[:num_prompts]
+
+
+def generate_cache_for_test_swap_decode_programs_for_cb(
+        model: str, prompts: list[str], parent_path: str):
+    '''
+    This function bakes the generation of prompts with long contexts. Which
+    currently are used in the test 
+    `test_spyre_cb::test_swap_decode_programs_for_cb`. 
+    '''
+
+    # Generate
+    assert len(prompts) == 4
+
+    p8k = 8 * 1024
+    p16k = 16 * 1024
+    p32k = 32 * 1024
+
+    import pickle
+    hf_outputs = generate_hf_output(model=model,
+                                    prompts=prompts[0:2],
+                                    max_new_tokens=p8k,
+                                    ignore_eos=True,
+                                    include_prompt=True)
+    with open(Path(parent_path) / 'prompts_8k_bs2.pickle', 'wb') as f:
+        f.write(pickle.dumps(hf_outputs))
+
+    hf_outputs = generate_hf_output(
+        model=model,
+        prompts=[prompts[2]],
+        max_new_tokens=p16k,
+        ignore_eos=True,
+        include_prompt=True,
+    )
+    with open(Path(parent_path) / 'prompts_16k_bs1.pickle', 'wb') as f:
+        f.write(pickle.dumps(hf_outputs))
+
+    hf_outputs = generate_hf_output(
+        model=model,
+        prompts=[prompts[3]],
+        max_new_tokens=p32k,
+        ignore_eos=True,
+        include_prompt=True,
+    )
+    with open(Path(parent_path) / 'prompts_32k_bs1.pickle', 'wb') as f:
+        f.write(pickle.dumps(hf_outputs))
